@@ -45,6 +45,7 @@
  * @CHAFA_CANVAS_MODE_INDEXED_256: 256 colors.
  * @CHAFA_CANVAS_MODE_INDEXED_240: 256 colors, but avoid using the lower 16 whose values vary between terminal environments.
  * @CHAFA_CANVAS_MODE_INDEXED_16: 16 colors using the aixterm ANSI extension.
+ * @CHAFA_CANVAS_MODE_INDEXED_8: 8 colors, compatible with original ANSI X3.64.
  * @CHAFA_CANVAS_MODE_FGBG_BGFG: Default foreground and background colors, plus inversion.
  * @CHAFA_CANVAS_MODE_FGBG: Default foreground and background colors. No ANSI codes will be used.
  * @CHAFA_CANVAS_MODE_MAX: Last supported canvas mode plus one.
@@ -76,6 +77,8 @@
  * ChafaPixelMode:
  * @CHAFA_PIXEL_MODE_SYMBOLS: Pixel data is approximated using character symbols ("ANSI art").
  * @CHAFA_PIXEL_MODE_SIXELS: Pixel data is encoded as sixels.
+ * @CHAFA_PIXEL_MODE_KITTY: Pixel data is encoded using the Kitty terminal protocol.
+ * @CHAFA_PIXEL_MODE_ITERM2: Pixel data is encoded using the iTerm2 terminal protocol.
  * @CHAFA_PIXEL_MODE_MAX: Last supported pixel mode plus one.
  **/
 
@@ -116,9 +119,13 @@ chafa_canvas_config_init (ChafaCanvasConfig *canvas_config)
     canvas_config->work_factor = 0.5;
     canvas_config->preprocessing_enabled = TRUE;
     canvas_config->optimizations = CHAFA_OPTIMIZATION_ALL;
+    canvas_config->fg_only_enabled = FALSE;
 
     chafa_symbol_map_init (&canvas_config->symbol_map);
-    chafa_symbol_map_add_by_tags (&canvas_config->symbol_map, CHAFA_SYMBOL_TAG_ALL);
+    chafa_symbol_map_add_by_tags (&canvas_config->symbol_map, CHAFA_SYMBOL_TAG_BLOCK);
+    chafa_symbol_map_add_by_tags (&canvas_config->symbol_map, CHAFA_SYMBOL_TAG_BORDER);
+    chafa_symbol_map_add_by_tags (&canvas_config->symbol_map, CHAFA_SYMBOL_TAG_SPACE);
+    chafa_symbol_map_remove_by_tags (&canvas_config->symbol_map, CHAFA_SYMBOL_TAG_WIDE);
 
     chafa_symbol_map_init (&canvas_config->fill_symbol_map);
 }
@@ -882,3 +889,50 @@ chafa_canvas_config_set_optimizations (ChafaCanvasConfig *config, ChafaOptimizat
 
     config->optimizations = optimizations;
 }
+
+/**
+ * chafa_canvas_config_get_fg_only_enabled:
+ * @config: A #ChafaCanvasConfig
+ *
+ * Queries whether to use foreground colors only, leaving the background
+ * unmodified in the canvas output. This is relevant only when the
+ * #ChafaPixelMode is set to #CHAFA_PIXEL_MODE_SYMBOLS.
+ *
+ * When this is set, the canvas will emit escape codes to set the foreground
+ * color only.
+ *
+ * Returns: %TRUE if using foreground colors only, %FALSE otherwise.
+ *
+ * Since: 1.8
+ **/
+gboolean
+chafa_canvas_config_get_fg_only_enabled (const ChafaCanvasConfig *config)
+{
+    g_return_val_if_fail (config != NULL, CHAFA_OPTIMIZATION_NONE);
+    g_return_val_if_fail (config->refs > 0, CHAFA_OPTIMIZATION_NONE);
+
+    return config->fg_only_enabled;
+}
+
+/**
+ * chafa_canvas_config_set_fg_only_enabled:
+ * @config: A #ChafaCanvasConfig
+ * @fg_only_enabled: Whether to use foreground colors only
+ *
+ * Indicates whether to use foreground colors only, leaving the background
+ * unmodified in the canvas output. This is relevant only when the
+ * #ChafaPixelMode is set to #CHAFA_PIXEL_MODE_SYMBOLS.
+ *
+ * When this is set, the canvas will emit escape codes to set the foreground
+ * color only.
+ *
+ * Since: 1.8
+ **/
+void
+chafa_canvas_config_set_fg_only_enabled (ChafaCanvasConfig *config, gboolean fg_only_enabled)
+{
+    g_return_if_fail (config != NULL);
+    g_return_if_fail (config->refs > 0);
+
+    config->fg_only_enabled = fg_only_enabled;
+}
