@@ -23,6 +23,7 @@
 #include <string.h>
 #include <glib.h>
 #include "chafa.h"
+#include "internal/chafa-batch.h"
 #include "internal/chafa-canvas-internal.h"
 #include "internal/chafa-canvas-printer.h"
 #include "internal/chafa-private.h"
@@ -270,8 +271,7 @@ eval_symbol (ChafaCanvas *canvas, ChafaWorkCell *wcell, gint sym_index,
 
     sym = &canvas->config.symbol_map.symbols [sym_index];
 
-    if (canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG
-        || canvas->config.fg_only_enabled)
+    if (canvas->config.fg_only_enabled)
     {
         eval.colors = canvas->default_colors;
     }
@@ -298,8 +298,7 @@ eval_symbol_wide (ChafaCanvas *canvas, ChafaWorkCell *wcell_a, ChafaWorkCell *wc
 
     sym2 = &canvas->config.symbol_map.symbols2 [sym_index];
 
-    if (canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG
-        || canvas->config.fg_only_enabled)
+    if (canvas->config.fg_only_enabled)
     {
         eval.colors = canvas->default_colors;
     }
@@ -346,7 +345,7 @@ pick_symbol_and_colors_slow (ChafaCanvas *canvas,
 
     g_assert (best_symbol >= 0);
 
-    if (canvas->config.fg_only_enabled)
+    if (canvas->extract_colors && canvas->config.fg_only_enabled)
         eval_symbol_colors (canvas, wcell, &canvas->config.symbol_map.symbols [best_symbol], &best_eval);
 
     *sym_out = canvas->config.symbol_map.symbols [best_symbol].c;
@@ -382,7 +381,7 @@ pick_symbol_and_colors_wide_slow (ChafaCanvas *canvas,
 
     g_assert (best_symbol >= 0);
 
-    if (canvas->config.fg_only_enabled)
+    if (canvas->extract_colors && canvas->config.fg_only_enabled)
         eval_symbol_colors_wide (canvas, wcell_a, wcell_b,
                                  &canvas->config.symbol_map.symbols2 [best_symbol].sym [0],
                                  &canvas->config.symbol_map.symbols2 [best_symbol].sym [1],
@@ -414,15 +413,13 @@ pick_symbol_and_colors_fast (ChafaCanvas *canvas,
 
     /* Generate short list of candidates */
 
-    if (canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG
-        || canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG_BGFG
-        || canvas->config.fg_only_enabled)
+    if (canvas->extract_colors && !canvas->config.fg_only_enabled)
     {
-        color_pair = canvas->default_colors;
+        chafa_work_cell_get_contrasting_color_pair (wcell, &color_pair);
     }
     else
     {
-        chafa_work_cell_get_contrasting_color_pair (wcell, &color_pair);
+        color_pair = canvas->default_colors;
     }
 
     bitmap = chafa_work_cell_to_bitmap (wcell, &color_pair);
@@ -430,9 +427,7 @@ pick_symbol_and_colors_fast (ChafaCanvas *canvas,
 
     chafa_symbol_map_find_candidates (&canvas->config.symbol_map,
                                       bitmap,
-                                      canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG
-                                      || canvas->config.fg_only_enabled
-                                      ? FALSE : TRUE,  /* Consider inverted? */
+                                      canvas->consider_inverted,
                                       candidates, &n_candidates);
 
     g_assert (n_candidates > 0);
@@ -451,7 +446,7 @@ pick_symbol_and_colors_fast (ChafaCanvas *canvas,
 
     g_assert (best_symbol >= 0);
 
-    if (canvas->config.fg_only_enabled)
+    if (canvas->extract_colors && canvas->config.fg_only_enabled)
         eval_symbol_colors (canvas, wcell, &canvas->config.symbol_map.symbols [best_symbol], &best_eval);
 
     *sym_out = canvas->config.symbol_map.symbols [best_symbol].c;
@@ -502,9 +497,7 @@ pick_symbol_and_colors_wide_fast (ChafaCanvas *canvas,
 
     chafa_symbol_map_find_wide_candidates (&canvas->config.symbol_map,
                                            bitmaps,
-                                           canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG
-                                           || canvas->config.fg_only_enabled
-                                           ? FALSE : TRUE,  /* Consider inverted? */
+                                           canvas->consider_inverted,
                                            candidates, &n_candidates);
 
     g_assert (n_candidates > 0);
@@ -524,7 +517,7 @@ pick_symbol_and_colors_wide_fast (ChafaCanvas *canvas,
 
     g_assert (best_symbol >= 0);
 
-    if (canvas->config.fg_only_enabled)
+    if (canvas->extract_colors && canvas->config.fg_only_enabled)
         eval_symbol_colors_wide (canvas, wcell_a, wcell_b,
                                  &canvas->config.symbol_map.symbols2 [best_symbol].sym [0],
                                  &canvas->config.symbol_map.symbols2 [best_symbol].sym [1],
@@ -542,14 +535,7 @@ pick_symbol_and_colors_wide_fast (ChafaCanvas *canvas,
 static const ChafaColor *
 get_palette_color_with_color_space (ChafaCanvas *canvas, gint index, ChafaColorSpace cs)
 {
-    if (index == CHAFA_PALETTE_INDEX_FG)
-        return &canvas->default_colors.colors [CHAFA_COLOR_PAIR_FG];
-    if (index == CHAFA_PALETTE_INDEX_BG)
-        return &canvas->default_colors.colors [CHAFA_COLOR_PAIR_BG];
-    if (index == CHAFA_PALETTE_INDEX_TRANSPARENT)
-        return &canvas->default_colors.colors [CHAFA_COLOR_PAIR_BG];
-
-    return chafa_get_palette_color_256 (index, cs);
+    return chafa_palette_get_color (&canvas->palette, cs, index);
 }
 
 static const ChafaColor *
@@ -683,9 +669,7 @@ apply_fill (ChafaCanvas *canvas, const ChafaWorkCell *wcell, ChafaCanvasCell *ce
     }
 
     chafa_symbol_map_find_fill_candidates (&canvas->config.fill_symbol_map, best_i,
-                                           canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG
-                                           || canvas->config.fg_only_enabled
-                                           ? FALSE : TRUE,  /* Consider inverted? */
+                                           canvas->consider_inverted,
                                            &sym_cand, &n_sym_cands);
 
     /* If we end up with a featureless symbol (space or fill), make
@@ -902,37 +886,26 @@ update_cells_row (ChafaCanvas *canvas, gint row)
     }
 }
 
-typedef struct
-{
-    gint row;
-}
-CellBuildWork;
-
 static void
-cell_build_worker (CellBuildWork *work, ChafaCanvas *canvas)
+cell_build_worker (ChafaBatchInfo *batch, ChafaCanvas *canvas)
 {
-    update_cells_row (canvas, work->row);
-    g_slice_free (CellBuildWork, work);
+    gint i;
+
+    for (i = 0; i < batch->n_rows; i++)
+    {
+        update_cells_row (canvas, batch->first_row + i);
+    }
 }
 
 static void
 update_cells (ChafaCanvas *canvas)
 {
-    GThreadPool *thread_pool = g_thread_pool_new ((GFunc) cell_build_worker,
-                                                  canvas,
-                                                  g_get_num_processors (),
-                                                  FALSE,
-                                                  NULL);
-    gint cy;
-
-    for (cy = 0; cy < canvas->config.height; cy++)
-    {
-        CellBuildWork *work = g_slice_new (CellBuildWork);
-        work->row = cy;
-        g_thread_pool_push (thread_pool, work, NULL);
-    }
-
-    g_thread_pool_free (thread_pool, FALSE, TRUE);
+    chafa_process_batches (canvas,
+                           (GFunc) cell_build_worker,
+                           NULL,  /* _post */
+                           canvas->config.height,
+                           chafa_get_n_actual_threads (),
+                           1);
 }
 
 static void
@@ -982,9 +955,7 @@ update_display_colors (ChafaCanvas *canvas)
      * We don't need to do this for monochrome modes, as they use the
      * FG/BG colors directly. */
 
-    if (canvas->config.fg_only_enabled
-        && canvas->config.canvas_mode != CHAFA_CANVAS_MODE_FGBG
-        && canvas->config.canvas_mode != CHAFA_CANVAS_MODE_FGBG_BGFG)
+    if (canvas->extract_colors && canvas->config.fg_only_enabled)
     {
         gint i;
 
@@ -1167,10 +1138,19 @@ chafa_canvas_new (const ChafaCanvasConfig *config)
 
     canvas->pixels = NULL;
     canvas->cells = g_new (ChafaCanvasCell, canvas->config.width * canvas->config.height);
-    canvas->work_factor_int = canvas->config.work_factor * 10 + 0.5;
+    canvas->work_factor_int = canvas->config.work_factor * 10 + 0.5f;
     canvas->needs_clear = TRUE;
     canvas->have_alpha = FALSE;
 
+    canvas->consider_inverted = !(canvas->config.fg_only_enabled
+                                  || canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG);
+
+    canvas->extract_colors = !(canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG
+                               || canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG_BGFG);
+
+    if (canvas->config.canvas_mode == CHAFA_CANVAS_MODE_FGBG)
+        canvas->config.fg_only_enabled = TRUE;
+
     chafa_symbol_map_prepare (&canvas->config.symbol_map);
     chafa_symbol_map_prepare (&canvas->config.fill_symbol_map);
 
@@ -1607,7 +1587,7 @@ chafa_canvas_get_char_at (ChafaCanvas *canvas, gint x, gint y)
  *
  * Sets the character at cell (x, y). The coordinates are zero-indexed. For
  * double-width characters, the leftmost cell must contain the character
- * and the rightmost cell must contain 0.
+ * and the cell to the right of it will automatically be set to 0.
  *
  * If the character is a nonprintable or zero-width, no change will be
  * made.
