@@ -94,6 +94,8 @@ static const SeqStr sixel_seqs [] =
 {
     { CHAFA_TERM_SEQ_BEGIN_SIXELS, "\033P%1;%2;%3q" },
     { CHAFA_TERM_SEQ_END_SIXELS, "\033\\" },
+    { CHAFA_TERM_SEQ_ENABLE_SIXEL_SCROLLING, "\033[?80l" },
+    { CHAFA_TERM_SEQ_DISABLE_SIXEL_SCROLLING, "\033[?80h" },
 
     { CHAFA_TERM_SEQ_MAX, NULL }
 };
@@ -227,8 +229,10 @@ detect_capabilities (ChafaTermInfo *ti, gchar **envp)
 {
     const gchar *term;
     const gchar *colorterm;
+    const gchar *konsole_version;
     const gchar *vte_version;
     const gchar *term_program;
+    const gchar *term_name;
     const gchar *tmux;
     const gchar *ctx_backend;
     const gchar *lc_terminal;
@@ -244,12 +248,18 @@ detect_capabilities (ChafaTermInfo *ti, gchar **envp)
     colorterm = g_environ_getenv (envp, "COLORTERM");
     if (!colorterm) colorterm = "";
 
+    konsole_version = g_environ_getenv (envp, "KONSOLE_VERSION");
+    if (!konsole_version) konsole_version = "";
+
     vte_version = g_environ_getenv (envp, "VTE_VERSION");
     if (!vte_version) vte_version = "";
 
     term_program = g_environ_getenv (envp, "TERM_PROGRAM");
     if (!term_program) term_program = "";
 
+    term_name = g_environ_getenv (envp, "TERMINAL_NAME");
+    if (!term_name) term_name = "";
+
     tmux = g_environ_getenv (envp, "TMUX");
     if (!tmux) tmux = "";
 
@@ -279,6 +289,13 @@ detect_capabilities (ChafaTermInfo *ti, gchar **envp)
             rep_seqs_local = rep_seqs;
     }
 
+    /* Konsole exports KONSOLE_VERSION */
+    if (strtoul (konsole_version, NULL, 10) >= 220370)
+    {
+        /* Konsole version 22.03.70+ supports sixel graphics */
+        gfx_seqs = sixel_seqs;
+    }
+
     /* The ctx terminal (https://ctx.graphics/) understands REP */
     if (strlen (ctx_backend) > 0)
         rep_seqs_local = rep_seqs;
@@ -307,6 +324,16 @@ detect_capabilities (ChafaTermInfo *ti, gchar **envp)
         gfx_seqs = iterm2_seqs;
     }
 
+    if (!strcasecmp (term_program, "WezTerm"))
+    {
+        gfx_seqs = sixel_seqs;
+    }
+
+    if (!strcasecmp (term_name, "contour"))
+    {
+        gfx_seqs = sixel_seqs;
+    }
+
     /* Apple Terminal sets TERM=xterm-256color, and does not support truecolor */
     if (!g_ascii_strcasecmp (term_program, "Apple_Terminal"))
         color_seq_list = color_256_list;
@@ -326,7 +353,7 @@ detect_capabilities (ChafaTermInfo *ti, gchar **envp)
         gfx_seqs = sixel_seqs;
     }
 
-    if (!strcmp(term, "foot") || !strcmp(term, "foot-direct"))
+    if (!strcmp (term, "foot") || !strncmp (term, "foot-", 5))
         gfx_seqs = sixel_seqs;
 
     /* rxvt 256-color really is 256 colors only */
