@@ -19,6 +19,7 @@
 
 #include "config.h"
 
+#include "internal/chafa-batch.h"
 #include "internal/chafa-pixops.h"
 #include "internal/smolscale/smolscale.h"
 
@@ -69,27 +70,15 @@ typedef struct
     gint have_alpha_int;
 
     Histogram hist;
-    gint n_batches_pixels;
-    gint n_rows_per_batch_pixels;
-
     SmolScaleCtx *scale_ctx;
 }
 PrepareContext;
 
 typedef struct
 {
-    gint first_row;
-    gint n_rows;
     Histogram hist;
 }
-PreparePixelsBatch1;
-
-typedef struct
-{
-    gint first_row;
-    gint n_rows;
-}
-PreparePixelsBatch2;
+PreparePixelsBatch1Ret;
 
 static gint
 rgb_to_intensity_fast (const ChafaColor *color)
@@ -224,10 +213,10 @@ fs_dither_grain (const ChafaDither *dither,
     gint grain_shift = dither->grain_width_shift + dither->grain_height_shift;
     ChafaColorAccum next_error = { 0 };
     ChafaColorAccum accum = { 0 };
-    ChafaColorCandidates cand = { 0 };
     ChafaPixel *p;
     ChafaColor acol;
     const ChafaColor *col;
+    gint index;
     gint x, y, i;
 
     p = pixel;
@@ -269,8 +258,8 @@ fs_dither_grain (const ChafaDither *dither,
     /* Don't try to dither alpha */
     acol.ch [3] = 0xff;
 
-    chafa_palette_lookup_nearest (palette, color_space, &acol, &cand);
-    col = chafa_palette_get_color (palette, color_space, cand.index [0]);
+    index = chafa_palette_lookup_nearest (palette, color_space, &acol, NULL);
+    col = chafa_palette_get_color (palette, color_space, index);
 
     for (i = 0; i < 3; i++)
     {
@@ -337,7 +326,7 @@ fs_dither (const ChafaDither *dither, const ChafaPalette *palette,
     dest_y >>= dither->grain_height_shift;
     n_rows >>= dither->grain_height_shift;
 
-    error_rows = alloca (width_grains * 2 * sizeof (ChafaColorAccum));
+    error_rows = g_malloc (width_grains * 2 * sizeof (ChafaColorAccum));
     error_row [0] = error_rows;
     error_row [1] = error_rows + width_grains;
 
@@ -381,7 +370,7 @@ fs_dither (const ChafaDither *dither, const ChafaPalette *palette,
         else
         {
             /* Backwards pass */
-            pixel = pixels + (y << dither->grain_height_shift) * (width + 1) - grain_width;
+            pixel = pixels + (y << dither->grain_height_shift) * width + width - grain_width;
 
             fs_dither_grain (dither, palette, color_space, pixel, width,
                              error_row [0] + width_grains - 1,
@@ -415,6 +404,8 @@ fs_dither (const ChafaDither *dither, const ChafaPalette *palette,
         error_row [0] = error_row [1];
         error_row [1] = pp;
     }
+
+    g_free (error_rows);
 }
 
 static void
@@ -445,7 +436,7 @@ fs_and_convert_rgb_to_din99d (const ChafaDither *dither, const ChafaPalette *pal
 }
 
 static void
-prepare_pixels_1_inner (PreparePixelsBatch1 *work,
+prepare_pixels_1_inner (PreparePixelsBatch1Ret *ret,
                         PrepareContext *prep_ctx,
                         const guint8 *data_p,
                         ChafaPixel *pixel_out,
@@ -472,13 +463,13 @@ prepare_pixels_1_inner (PreparePixelsBatch1 *work,
     if (col->ch [3] > 127)
     {
         gint v = rgb_to_intensity_fast (col);
-        work->hist.c [v]++;
-        work->hist.n_samples++;
+        ret->hist.c [v]++;
+        ret->hist.n_samples++;
     }
 }
 
 static void
-prepare_pixels_1_worker_nearest (PreparePixelsBatch1 *work, PrepareContext *prep_ctx)
+prepare_pixels_1_worker_nearest (ChafaBatchInfo *batch, PrepareContext *prep_ctx)
 {
     ChafaPixel *pixel;
     gint dest_y;
@@ -488,10 +479,14 @@ prepare_pixels_1_worker_nearest (PreparePixelsBatch1 *work, PrepareContext *prep
     const guint8 *data;
     gint n_rows;
     gint rowstride;
+    PreparePixelsBatch1Ret *ret;
 
-    dest_y = work->first_row;
+    ret = g_new (PreparePixelsBatch1Ret, 1);
+    batch->ret_p = ret;
+
+    dest_y = batch->first_row;
     data = prep_ctx->src_pixels;
-    n_rows = work->n_rows;
+    n_rows = batch->n_rows;
     rowstride = prep_ctx->src_rowstride;
 
     x_inc = (prep_ctx->src_width * FIXED_MULT) / (prep_ctx->dest_width);
@@ -508,7 +503,7 @@ prepare_pixels_1_worker_nearest (PreparePixelsBatch1 *work, PrepareContext *prep
         for (px = 0; px < prep_ctx->dest_width; px++)
         {
             const guint8 *data_p = data_row_p + ((px * x_inc) / FIXED_MULT) * 4;
-            prepare_pixels_1_inner (work, prep_ctx, data_p, pixel++, &alpha_sum);
+            prepare_pixels_1_inner (ret, prep_ctx, data_p, pixel++, &alpha_sum);
         }
     }
 
@@ -517,23 +512,27 @@ prepare_pixels_1_worker_nearest (PreparePixelsBatch1 *work, PrepareContext *prep
 }
 
 static void
-prepare_pixels_1_worker_smooth (PreparePixelsBatch1 *work, PrepareContext *prep_ctx)
+prepare_pixels_1_worker_smooth (ChafaBatchInfo *batch, PrepareContext *prep_ctx)
 {
     ChafaPixel *pixel, *pixel_max;
     gint alpha_sum = 0;
     guint8 *scaled_data;
     const guint8 *data_p;
+    PreparePixelsBatch1Ret *ret;
 
-    scaled_data = g_malloc (prep_ctx->dest_width * work->n_rows * sizeof (guint32));
-    smol_scale_batch_full (prep_ctx->scale_ctx, scaled_data, work->first_row, work->n_rows);
+    ret = g_new0 (PreparePixelsBatch1Ret, 1);
+    batch->ret_p = ret;
+
+    scaled_data = g_malloc (prep_ctx->dest_width * batch->n_rows * sizeof (guint32));
+    smol_scale_batch_full (prep_ctx->scale_ctx, scaled_data, batch->first_row, batch->n_rows);
 
     data_p = scaled_data;
-    pixel = prep_ctx->dest_pixels + work->first_row * prep_ctx->dest_width;
-    pixel_max = pixel + work->n_rows * prep_ctx->dest_width;
+    pixel = prep_ctx->dest_pixels + batch->first_row * prep_ctx->dest_width;
+    pixel_max = pixel + batch->n_rows * prep_ctx->dest_width;
 
     while (pixel < pixel_max)
     {
-        prepare_pixels_1_inner (work, prep_ctx, data_p, pixel++, &alpha_sum);
+        prepare_pixels_1_inner (ret, prep_ctx, data_p, pixel++, &alpha_sum);
         data_p += 4;
     }
 
@@ -543,13 +542,23 @@ prepare_pixels_1_worker_smooth (PreparePixelsBatch1 *work, PrepareContext *prep_
         g_atomic_int_set (&prep_ctx->have_alpha_int, 1);
 }
 
+static void
+pass_1_post (ChafaBatchInfo *batch, PrepareContext *prep_ctx)
+{
+    PreparePixelsBatch1Ret *ret = batch->ret_p;
+
+    if (prep_ctx->preprocessing_enabled)
+    {
+        sum_histograms (&ret->hist, &prep_ctx->hist);
+    }
+
+    g_free (ret);
+}
+
 static void
 prepare_pixels_pass_1 (PrepareContext *prep_ctx)
 {
-    GThreadPool *thread_pool;
-    PreparePixelsBatch1 *batches;
-    gint cy;
-    gint i;
+    GFunc batch_func;
 
     /* First pass
      * ----------
@@ -560,38 +569,21 @@ prepare_pixels_pass_1 (PrepareContext *prep_ctx)
      * - Figure out if we have alpha transparency
      */
 
-    batches = g_new0 (PreparePixelsBatch1, prep_ctx->n_batches_pixels);
+    batch_func = (GFunc) ((prep_ctx->work_factor_int < 3
+                           && prep_ctx->src_pixel_type == CHAFA_PIXEL_RGBA8_UNASSOCIATED)
+                          ? prepare_pixels_1_worker_nearest
+                          : prepare_pixels_1_worker_smooth);
 
-    thread_pool = g_thread_pool_new ((GFunc) ((prep_ctx->work_factor_int < 3
-                                               && prep_ctx->src_pixel_type == CHAFA_PIXEL_RGBA8_UNASSOCIATED)
-                                              ? prepare_pixels_1_worker_nearest
-                                              : prepare_pixels_1_worker_smooth),
-                                     prep_ctx,
-                                     g_get_num_processors (),
-                                     FALSE,
-                                     NULL);
-
-    for (cy = 0, i = 0;
-         cy < prep_ctx->dest_height;
-         cy += prep_ctx->n_rows_per_batch_pixels, i++)
-    {
-        PreparePixelsBatch1 *batch = &batches [i];
-
-        batch->first_row = cy;
-        batch->n_rows = MIN (prep_ctx->dest_height - cy, prep_ctx->n_rows_per_batch_pixels);
-
-        g_thread_pool_push (thread_pool, batch, NULL);
-    }
-
-    /* Wait for threads to finish */
-    g_thread_pool_free (thread_pool, FALSE, TRUE);
+    chafa_process_batches (prep_ctx,
+                           (GFunc) batch_func,
+                           (GFunc) pass_1_post,
+                           prep_ctx->dest_height,
+                           chafa_get_n_actual_threads (),
+                           1);
 
     /* Generate final histogram */
     if (prep_ctx->preprocessing_enabled)
     {
-        for (i = 0; i < prep_ctx->n_batches_pixels; i++)
-            sum_histograms (&batches [i].hist, &prep_ctx->hist);
-
         switch (prep_ctx->palette_type)
         {
           case CHAFA_PALETTE_TYPE_FIXED_16:
@@ -605,8 +597,6 @@ prepare_pixels_pass_1 (PrepareContext *prep_ctx)
             break;
         }
     }
-
-    g_free (batches);
 }
 
 static void
@@ -627,19 +617,19 @@ composite_alpha_on_bg (ChafaColor bg_color,
 }
 
 static void
-prepare_pixels_2_worker (PreparePixelsBatch2 *work, PrepareContext *prep_ctx)
+prepare_pixels_2_worker (ChafaBatchInfo *batch, PrepareContext *prep_ctx)
 {
     if (prep_ctx->preprocessing_enabled
         && (prep_ctx->palette_type == CHAFA_PALETTE_TYPE_FIXED_16
             ||prep_ctx->palette_type == CHAFA_PALETTE_TYPE_FIXED_8
             || prep_ctx->palette_type == CHAFA_PALETTE_TYPE_FIXED_FGBG))
         normalize_rgb (prep_ctx->dest_pixels, &prep_ctx->hist, prep_ctx->dest_width,
-                       work->first_row, work->n_rows);
+                       batch->first_row, batch->n_rows);
 
     if (prep_ctx->have_alpha_int)
         composite_alpha_on_bg (prep_ctx->bg_color_rgb,
                                prep_ctx->dest_pixels, prep_ctx->dest_width,
-                               work->first_row, work->n_rows);
+                               batch->first_row, batch->n_rows);
 
     if (prep_ctx->color_space == CHAFA_COLOR_SPACE_DIN99D)
     {
@@ -648,8 +638,8 @@ prepare_pixels_2_worker (PreparePixelsBatch2 *work, PrepareContext *prep_ctx)
             bayer_and_convert_rgb_to_din99d (prep_ctx->dither,
                                              prep_ctx->dest_pixels,
                                              prep_ctx->dest_width,
-                                             work->first_row,
-                                             work->n_rows);
+                                             batch->first_row,
+                                             batch->n_rows);
         }
         else if (prep_ctx->dither->mode == CHAFA_DITHER_MODE_DIFFUSION)
         {
@@ -657,15 +647,15 @@ prepare_pixels_2_worker (PreparePixelsBatch2 *work, PrepareContext *prep_ctx)
                                           prep_ctx->palette,
                                           prep_ctx->dest_pixels,
                                           prep_ctx->dest_width,
-                                          work->first_row,
-                                          work->n_rows);
+                                          batch->first_row,
+                                          batch->n_rows);
         }
         else
         {
             convert_rgb_to_din99d (prep_ctx->dest_pixels,
                                    prep_ctx->dest_width,
-                                   work->first_row,
-                                   work->n_rows);
+                                   batch->first_row,
+                                   batch->n_rows);
         }
     }
     else if (prep_ctx->dither->mode == CHAFA_DITHER_MODE_ORDERED)
@@ -673,8 +663,8 @@ prepare_pixels_2_worker (PreparePixelsBatch2 *work, PrepareContext *prep_ctx)
         bayer_dither (prep_ctx->dither,
                       prep_ctx->dest_pixels,
                       prep_ctx->dest_width,
-                      work->first_row,
-                      work->n_rows);
+                      batch->first_row,
+                      batch->n_rows);
     }
     else if (prep_ctx->dither->mode == CHAFA_DITHER_MODE_DIFFUSION)
     {
@@ -683,8 +673,8 @@ prepare_pixels_2_worker (PreparePixelsBatch2 *work, PrepareContext *prep_ctx)
                    prep_ctx->color_space,
                    prep_ctx->dest_pixels,
                    prep_ctx->dest_width,
-                   work->first_row,
-                   work->n_rows);
+                   batch->first_row,
+                   batch->n_rows);
     }
 }
 
@@ -706,10 +696,8 @@ need_pass_2 (PrepareContext *prep_ctx)
 static void
 prepare_pixels_pass_2 (PrepareContext *prep_ctx)
 {
-    GThreadPool *thread_pool;
-    PreparePixelsBatch1 *batches;
-    gint cy;
-    gint i;
+    gint n_batches;
+    gint batch_unit = 1;
 
     /* Second pass
      * -----------
@@ -722,30 +710,23 @@ prepare_pixels_pass_2 (PrepareContext *prep_ctx)
     if (!need_pass_2 (prep_ctx))
         return;
 
-    batches = g_new0 (PreparePixelsBatch1, prep_ctx->n_batches_pixels);
+    n_batches = chafa_get_n_actual_threads ();
 
-    thread_pool = g_thread_pool_new ((GFunc) prepare_pixels_2_worker,
-                                     prep_ctx,
-                                     g_get_num_processors (),
-                                     FALSE,
-                                     NULL);
-
-    for (cy = 0, i = 0;
-         cy < prep_ctx->dest_height;
-         cy += prep_ctx->n_rows_per_batch_pixels, i++)
+    /* Floyd-Steinberg diffusion needs the batch size to be a multiple of the
+     * grain height. It also needs to run in a single thread to propagate the
+     * quantization error correctly. */
+    if (prep_ctx->dither->mode == CHAFA_DITHER_MODE_DIFFUSION)
     {
-        PreparePixelsBatch1 *batch = &batches [i];
-
-        batch->first_row = cy;
-        batch->n_rows = MIN (prep_ctx->dest_height - cy, prep_ctx->n_rows_per_batch_pixels);
-
-        g_thread_pool_push (thread_pool, batch, NULL);
+        n_batches = 1;
+        batch_unit = 1 << prep_ctx->dither->grain_height_shift;
     }
 
-    /* Wait for threads to finish */
-    g_thread_pool_free (thread_pool, FALSE, TRUE);
-
-    g_free (batches);
+    chafa_process_batches (prep_ctx,
+                           (GFunc) prepare_pixels_2_worker,
+                           NULL,  /* _post */
+                           prep_ctx->dest_height,
+                           n_batches,
+                           batch_unit);
 }
 
 void
@@ -764,9 +745,6 @@ chafa_prepare_pixel_data_for_symbols (const ChafaPalette *palette,
                                       gint dest_height)
 {
     PrepareContext prep_ctx = { 0 };
-    guint n_cpus;
-
-    n_cpus = g_get_num_processors ();
 
     prep_ctx.palette = palette;
     prep_ctx.dither = dither;
@@ -789,9 +767,6 @@ chafa_prepare_pixel_data_for_symbols (const ChafaPalette *palette,
     prep_ctx.dest_width = dest_width;
     prep_ctx.dest_height = dest_height;
 
-    prep_ctx.n_batches_pixels = (prep_ctx.dest_height + n_cpus - 1) / n_cpus;
-    prep_ctx.n_rows_per_batch_pixels = (prep_ctx.dest_height + prep_ctx.n_batches_pixels - 1) / prep_ctx.n_batches_pixels;
-
     prep_ctx.scale_ctx = smol_scale_new ((SmolPixelType) prep_ctx.src_pixel_type,
                                          (const guint32 *) prep_ctx.src_pixels,
                                          prep_ctx.src_width,
